'use strict'

const TimedCache = require('./timed-cache')

exports.create = create

const lodash = require('lodash')

// In seconds
const DEFAULT_DELAY = 30 // 30 seconds
const DEFAULT_SNOOZE = 60 * 5 // 5 minutes
const BLOCKED_THRESHOLD = 60 * 1000 // 1 minute

function create () {
  return new ActionCache()
}

class ActionCache {
  constructor () {
    this._state = new Map()
    this._blocked = new TimedCache(BLOCKED_THRESHOLD)
  }

  shouldRun (id, params) {
    if (!params.view || !params.agents) {
      throw new Error('Can not run shouldRun without view and agents.')
    }

    let state = this.getState(id, params)
    params.state = state
    return !this.isDelayed(params) && !this.isSnoozed(params)
  }

  isDelayed (params) {
    if (!params.state || !params.state.enter || !params.now) {
      throw new Error('Required params: state, state.enter, and now')
    }

    const delay = (params.view.delay || DEFAULT_DELAY) * 1000
    return params.now - params.state.enter <= delay
  }

  isSnoozed (params) {
    if (!params.state || !params.now) {
      throw new Error('Required params: state and now')
    }

    const last = params.state.last
    const snooze = (params.view.snooze || DEFAULT_SNOOZE) * 1000
    if (!last) return false
    return params.now - last <= snooze
  }

  setLastRun (id, params) {
    let state = this.getState(id, params)
    state.last = params.now
    this._state.set(id, state)
  }

  setEnter (id, params) {
    if (!params.now) {
      throw new Error('Can not set enter without now value.')
    }

    let state = this.getState(id, params)
    if (!state.enter) {
      state.enter = params.now
      this._state.set(id, state)
    }
  }

  setEmpty (id, params) {
    let state = this.getState(id, params)
    state.enter = false
    this._state.set(id, state)
  }

  getState (id, params) {
    if (!id) {
      throw new Error('Can not look up action state without id.')
    }

    let state = this._state.get(id)
    if (!state) {
      state = {
        enter: false,
        last: false
      }
      this._state.set(id, state)

      if (params.agents) {
        state.enter = params.now
        state.last = false
      }
    }

    return state
  }

  // addr is a slack or webhook url or an email address
  shouldVulnRun (params) {
    if (!params.vuln || !params.vuln.id) {
      throw new Error('Can not check vuln notification state without vuln id.')
    }

    if (!params.agent || !params.agent.id) {
      throw new Error('Can not check vuln notification state without agent id.')
    }

    if (!params.addr) {
      throw new Error('Can not check vuln notification state without notification address.')
    }

    let entry = this._state.get(params.vuln.id)
    if (!entry) entry = {}

    return !lodash.get(entry, `${params.agent.id}:${params.addr}`)
  }

  deleteState (params) {
    this._state.delete(params.view.id)
  }
}
