'use strict'

const fs = require('fs')
const path = require('path')

const async = require('async')
const lodash = require('lodash')

exports.run = run
exports.getEntries = getEntries

// Deletes assets exceeding limits in config.  Also deletes directories
// that are empty.
// Note, directories that become empty because files get deleted won't
// be deleted during this run - but presumably will for the next run.
// Note, currently does not update the asset index, delegating to the
// disk-asset-manager's validation check to do that.
function run (context, cb) {
  const assetsDir = path.join(context.config.data.dir, 'assets')

  // get files and directories under assetsDir
  getEntries(assetsDir, gotEntries)

  // handle files and directories returned
  function gotEntries (err, entries) {
    if (err) return cb(err)

    // remove directories from entries
    const dirs = lodash.remove(entries, (entry) => entry.isDir)

    // build list of empty directories
    const emptyDirs = dirs
      .filter((entry) => entry.children === 0)
      .map((entry) => entry.absName)

    // process files
    processFiles(context, assetsDir, entries, (err) => {
      if (err) return cb(err)

      // delete empty directories
      async.each(emptyDirs, fs.rmdir, cb)
    })
  }
}

// Process files found in assets directory, deleting files to fit in
// configured limits of count and total size.
function processFiles (context, assetsDir, files, cb) {
  const indexFile = path.join(assetsDir, 'index.json')
  lodash.remove(files, (el) => el.absName === indexFile)

  context.settingsManager.getSettings('generalSettings', (err, settings) => {
    if (err) throw err

    const keepProfiles = lodash.get(settings, ['generalSettingsMain', 'keepProfiles'])
    const keepSnapshots = lodash.get(settings, ['generalSettingsMain', 'keepSnapshots'])

    const maxAssets = context.config.assets.maxAssets
    const maxAssetsSize = context.config.assets.maxAssetsMb * 1000 * 1000

    // sort files by date, oldest at end
    files = lodash.sortBy(files, (file) => -file.mtime)

    // collects files to be deleted
    const toDelete = []

    // Create a map of paths to objects from the cache
    const pathToAsset = new Map()
    for (let asset of context.diskAssetManager._assets.values()) {
      const fullPath = path.join(assetsDir, asset.fileName) + (asset.compressed ? '.gz' : '')
      pathToAsset.set(fullPath, asset)
    }

    const profileCutoff = keepProfiles ? Date.now() - (keepProfiles * 24 * 60 * 60 * 1000) : 0
    const snapshotCutoff = keepSnapshots ? Date.now() - (keepSnapshots * 24 * 60 * 60 * 1000) : 0

    // Starred assets are exempt from GC, and do not count towards space quota
    files = files.filter(path => {
      const asset = pathToAsset.get(path.absName)

      // No index means eligible for gc
      if (!asset) return true

      // User-requested delete comes before other considerations
      if (asset.deleteSoon && Date.now() > asset.deleteSoon) {
        toDelete.push(path)
        // Already added to set, remove from the list
        return false
      }

      // Starred asset is not automatically deleted
      if (asset.starred) return false

      // Process the UI-configured retention settings
      if (asset.type === 'profile' && profileCutoff && asset.time < profileCutoff) {
        toDelete.push(path)
        // Already added to set, remove from the list
        return false
      }
      if ((asset.type === 'snapshot' || asset.type === 'snapshot-summary') && snapshotCutoff && asset.time < snapshotCutoff) {
        toDelete.push(path)
        // Already added to set, remove from the list
        return false
      }

      // Ok to gc by size + count
      return true
    })

    // mark for delete files to fit in count constraint
    while (files.length > maxAssets) {
      toDelete.push(files.pop())
    }

    // get current size of all files
    const reducer = (acc, file) => acc + file.size
    let currSize = files.reduce(reducer, 0)

    // mark for delete files to fit in size constraint
    while (currSize > maxAssetsSize) {
      const file = files.pop()
      toDelete.push(file)
      currSize -= file.size
    }

    // delete the files
    async.each(
      toDelete,
      (file, asyncCB) => fs.unlink(file.absName, asyncCB),
      (err) => cb(err)
    )
  })
}

// Return an array of file objects, for files under specified dir, recursively:
//    obj.absName  - absolute name of the file/dir
//    obj.isDir    - true for directories
//    obj.mtime    - (files) epoch time integer of modification time
//    obj.size     - (files) size of file in bytes
//    obj.children - (directories) number of children in directory
function getEntries (dir, cb) {
  const absName = path.resolve(dir)

  const queue = []
  const entries = []

  queue.push(absName)

  async.until(
    () => queue.length === 0,
    (cb) => processFile(queue, entries, queue.pop(), cb),
    (err) => {
      if (err) return cb(err)
      cb(null, entries)
    }
  )
}

// process a file found during processing of getEntries()
function processFile (queue, entries, absName, cb) {
  fs.stat(absName, (err, stats) => {
    if (err) {
      // for missing files, nothing to do!
      if (err.code === 'ENOENT') return cb()
      return cb(err)
    }

    // if file, add entry
    if (stats.isFile()) {
      entries.push({
        absName: absName,
        isDir: false,
        mtime: stats.mtime.getTime(),
        size: stats.size
      })
      return cb()
    }

    // if not dir, nothing else to do
    if (!stats.isDirectory()) {
      return cb()
    }

    // if dir, get children, queue for processing
    fs.readdir(absName, (err, files) => {
      if (err) return cb(err)

      entries.push({
        absName: absName,
        isDir: true,
        children: files.length
      })

      for (let file of files) {
        queue.push(path.join(absName, file))
      }

      return cb()
    })
  })
}
