'use strict'

const lodash = require('lodash')

const util = require('../util')

exports.syncPollTillDone = syncPollTillDone
exports.getHandlerForCommandWithAsset = getHandlerForCommandWithAsset
exports.getBroadcastFilter = getBroadcastFilter
exports.updateSettingByType = updateSettingByType
exports.deleteSettingByType = deleteSettingByType
exports.runAssetCommand = runAssetCommand
exports.endProfile = endProfile
exports.cancelProfile = cancelProfile
exports.normalizeAsset = normalizeAsset

const Logger = util.getLogger(__dirname, __filename)

const BroadcastFilterQParms = 'app,hostname,tags'.split(/,/g)

// Run fnPoll every interval ms, until Date.now > end or eventStream is closed.
// When finished, call fnDone.
// First run will be after the specified interval.
// It's assumed the fnPoll function is synchronous, so that results of
// calling it multiple times won't result in out-of-order responses.
function syncPollTillDone (eventStream, interval, endDate, fnPoll, fnDone) {
  setTimeout(
    () => _syncPollTillDone(eventStream, interval, endDate, fnPoll, fnDone),
    interval
  ).unref()
}

function _syncPollTillDone (eventStream, interval, endDate, fnPoll, fnDone) {
  let done = false
  if (eventStream.isClosed()) done = true
  if (endDate && Date.now() > endDate) done = true

  if (done) {
    if (fnDone) fnDone()
    return
  }

  fnPoll()

  setTimeout(
    () => _syncPollTillDone(eventStream, interval, endDate, fnPoll, fnDone),
    interval
  ).unref()
}

// Returns an HTTP request/response handler for the given context and command.
function getHandlerForCommandWithAsset (command, getCommandArgs) {
  return (req, res) =>
    handlerForCommandWithAsset(req, res, command, getCommandArgs)
}

// HTTP request/response handler for the given context and command.
function handlerForCommandWithAsset (req, res, command, getCommandArgs) {
  const context = req.nsConsole.context
  const qData = req.nsConsole.qData
  let agentId = qData.id

  var agent
  if ((agent = context.agentManager.byId(agentId)) == null) {
    return sendCommandErrorResponse(context, res, {
      status: 404,
      command: command,
      error: {message: 'agent not found'},
      agentId: agentId
    })
  }

  // Fully qualify id
  agentId = agent.id

  const commandObject = {
    command: command,
    id: agentId,
    args: {
      metadata: { reason: 'User' }
    }
  }

  if (getCommandArgs != null) {
    Object.assign(commandObject.args, getCommandArgs(req))
  }

  const requestId = runAssetCommand(commandObject, context, (err, asset) => {
    if (err) {
      return sendCommandErrorResponse(context, res, {
        status: 500,
        command: command,
        error: {message: err.message},
        agentId: agentId,
        requestId: requestId,
        asset: asset
      })
    }

    Logger.debug(`asset complete for requestId: ${requestId}`, err)
    sendAsset(asset)
  })

  // Send the asset info as the http response
  function sendAsset (asset) {
    const assetURL = `/api/v3/asset/${asset.bulkId}`

    // the response could have been sent already if the client side socket
    // closed (user cancelled)
    if (res.headersSent) return

    res.location(assetURL)
    res.status(201)

    const event = {
      id: agentId,
      time: new Date(),
      asset: asset.bulkId,
      type: asset.type,
      assetURL: assetURL,
      duration: asset.duration
    }

    if (asset.summaryAsset) {
      event.summaryAsset = asset.summaryAsset
    }

    const agent = context.agentManager.byId(agentId)
    if (agent && agent.info) {
      event.app = agent.info.app
      event.tags = agent.info.tags
      event.hostname = agent.info.hostname
      event.pid = agent.info.pid
      event.title = agent.metrics ? agent.metrics.title : undefined
    }

    res.send(event)
  }
}

function runAssetCommand (commandObject, context, _cb) {
  const cb = lodash.once((err, asset) => {
    cleanUpCommandRequest(context, requestId, asset)
    _cb(err, asset)
  })
  // Send the command.
  Logger.debug('sending agent command:', commandObject)

  const profileMinDuration = commandObject.profileMinDuration
  delete commandObject.profileMinDuration

  const requestId = context.commandDispatcher.send(commandObject)

  // On a command error, send appropriate response.
  context.commandDispatcher.on(`error:${requestId}`, cb)

  context.commandDispatcher.on(`data:${requestId}`, cmdResponse => {
    Logger.debug(`command received for ${requestId}`, cmdResponse)
  })

  // On a command success, set up the asset handling.
  context.assetManager.once(`add:${requestId}`, asset => {
    Logger.debug(`asset added, waiting for complete: ${asset}`)

    // For older agents that can't proxy metadata, copy it to the asset object
    const viewId = lodash.get(commandObject, ['args', 'metadata', 'viewId'])
    if (viewId && !asset.viewId) {
      asset.viewId = viewId
    }

    asset.once('complete', err => {
      // file should now be written, but wait for next tick
      // to make sure bits are flushed out right (have seen
      // further processing not read the end of the file)
      setImmediate(cb, err, asset)
    })
  })

  context.eventManager.emitSystemEvent('asset-initiated', commandObject.id, {
    type: commandObject.command,
    requestId,
    profileMinDuration
  })

  return requestId
}

function endProfile (context, requestId, agentId) {
  Logger.debug(`ending profile ${requestId} on ${agentId}`)
  context.commandDispatcher.send({
    command: 'profile_stop',
    id: agentId,
    args: {
      stopRequestId: requestId
    }
  })

  context.commandDispatcher.on(`error:${requestId}`, err => {
    Logger.warn(`Error stopping profile ${requestId}: ${err.message}`)
  })
}

function cancelProfile (context, requestId, agentId) {
  Logger.debug(`canceling profile ${requestId} on ${agentId}`)
  context.commandDispatcher.send({
    command: 'profile',
    id: agentId,
    cancel: true,
    requestId
  })

  context.commandDispatcher.on(`error:${requestId}`, err => {
    Logger.warn(`Error canceling profile ${requestId}: ${err.message}`)
  })

  context.eventManager.emitSystemEvent('asset-canceled', agentId, {
    type: 'profile',
    requestId
  })
}

// Send an error response
function sendCommandErrorResponse (context, res, parms) {
  const status = parms.status || 500
  const command = parms.command || 'unknown command'
  const error = parms.error || 'unknown error'
  const agentId = parms.agentId
  const requestId = parms.requestId
  const asset = parms.asset

  Logger.warn(`error sending command ${command}:`, error)

  if (res.__nsAlreadySent) return
  res.__nsAlreadySent = true

  cleanUpCommandRequest(context, requestId, asset)

  const body = {}

  if (agentId) body.id = agentId

  body.time = new Date()
  body.error = error

  res.status(status).send(body)
}

// Cleanup all the command handlers etc we registered
function cleanUpCommandRequest (context, requestId, asset) {
  if (requestId != null) {
    context.commandDispatcher.removeAllListeners(`end:${requestId}`)
    context.commandDispatcher.removeAllListeners(`data:${requestId}`)
    context.commandDispatcher.removeAllListeners(`error:${requestId}`)
  }

  if (asset != null) {
    asset.removeAllListeners('complete')

    const assetId = asset.getAssetId()
    if (assetId != null) {
      context.assetManager.removeAllListeners(`add:${requestId}`)
    }
  }
}

// Return an agent message filter from agentScope querystring data
function getBroadcastFilter (qData) {
  const result = lodash.pick(qData, BroadcastFilterQParms)

  if (lodash.isEmpty(result)) return null

  return result
}

function updateSettingByType (req, res, key, cb) {
  const val = req.body
  const settingsManager = req.nsConsole.context.settingsManager

  // check if body parser set an error
  if (req.nsConsole.bodyErr) {
    return errorReturn(res, 400, 'invalid JSON request body')
  }

  // update the settings with settingsManager
  settingsManager.setSetting(key, val, (err, item) => {
    if (err) {
      Logger.warn(err)
      return errorReturn(res, 400, 'error updating settings')
    }

    res.status(200)
    res.send(item)
  })
}

function deleteSettingByType (req, res, key, cb) {
  // check if body parser set an error
  if (req.nsConsole.bodyErr) {
    return errorReturn(res, 400, 'invalid JSON request body')
  }

  const settingsManager = req.nsConsole.context.settingsManager
  const id = req.query.id

  // update the settings with settingsManager
  settingsManager.deleteSetting(key, id, (err) => {
    if (err) {
      Logger.warn(err)
      return errorReturn(res, 400, 'error deleting settings')
    }

    res.status(200)
    res.send({time: Date.now()})
  })
}

// Standard processing for an error return.
function errorReturn (res, status, message) {
  res.status(status)
  res.send({error: {
    message: message
  }})
}

// Return an asset object without the fileName property
function normalizeAsset (asset) {
  asset = Object.assign({}, asset)
  delete asset.fileName
  asset.time = new Date(asset.time)
  // turn undefined to false for sortability
  asset.starred = !!asset.starred
  return asset
}
