'use strict'

exports.handleRequest = handleRequest

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsConsole.context
  const qData = req.nsConsole.qData

  Logger.debug('handling request')

  const summary = getSummary(context.agentManager, qData.q)
  res.status(200)
  res.send(summary)
}

// get summary info of agents which match the query
function getSummary (agentManager, queryExpression) {
  let count = 0
  const apps = new Counter()
  const tags = new Counter()
  let rss = 0
  let cpuPercent = 0
  const node = new Counter()
  const nsolid = new Counter()
  const vulns = new Counter()
  const hiddenVulns = new Counter()
  let packages = 0

  const agents = agentManager.query(queryExpression)

  agents.forEach((agent) => {
    count++
    if (agent.info != null) {
      apps.increment(agent.info.app)
      agent.info.tags.forEach((tag) => {
        tags.increment(tag)
      })
      if (agent.info.versions != null) {
        node.increment(agent.info.versions.node)
        nsolid.increment(agent.info.versions.nsolid)
      }
    }
    if (agent.metrics != null) {
      rss += agent.metrics.rss
      cpuPercent += agent.metrics.cpuPercent
    }

    if (agent.packages != null) {
      packages += agent.packages.length
    }

    if (agent.vulns != null) {
      for (let vuln of agent.vulns) {
        if (vuln.nsolidMetaData.hidden) {
          hiddenVulns.increment(vuln.id)
        } else {
          vulns.increment(vuln.id)
        }
      }
    }
  })

  const summary = {
    time: new Date(),
    processes: count,
    apps: apps.contents(),
    tags: tags.contents(),
    rss: rss,
    cpu: cpuPercent,
    node: node.contents(),
    nsolid: nsolid.contents(),
    packages: packages,
    vulnerabilities: vulns.contents(),
    hiddenVulnerabilities: hiddenVulns.contents()
  }

  return summary
}

class Counter {
  constructor () {
    this._store = {}
  }

  increment (key) {
    let entry = this._store[key]
    if (entry == null) {
      this._store[key] = 1
    } else {
      this._store[key] = entry + 1
    }
  }

  contents () {
    return Object.assign(this._store)
  }
}
