'use strict'

const fs = require('fs')
const path = require('path')
const childProcess = require('child_process')

const lodash = require('lodash')
const portFinder = require('portfinder')

const api = require('./api')
const util = require('../util')

const createSchema = require('./schema').create

exports.launch = launch

const Logger = util.getLogger(__dirname, __filename)

// Launch an instance of InfluxDB.
function launch (context, cb) {
  const config = context.config

  if (!config.influxdb) config.influxdb = {}

  // if using external InfluxDB, just call createSchema and return
  if (config.influxdb.url) {
    Logger.info(`Connecting to influxdb at ${config.influxdb.url}`)
    context.usingExternalInflux = true

    createSchema(context, err => {
      if (err) {
        Logger.error(err, 'error creating InfluxDB schema')
        // Do not propagate this error, so startup continues.  It was decided
        // that an error banner on the console is a better signal than the
        // console failing to start, if external influxdb has connection errors
      }
      cb()
    })
    return
  }

  config.influxdb.host = 'localhost'
  config.influxdb.metaDir = path.join(config.data.dir, 'influxdb', 'meta')
  config.influxdb.dataDir = path.join(config.data.dir, 'influxdb', 'data')
  config.influxdb.walDir = path.join(config.data.dir, 'influxdb', 'wal')

  let err

  err = util.ensureDirExists(config.influxdb.metaDir)
  if (err) return util.cbNextTick(cb, err)

  err = util.ensureDirExists(config.influxdb.dataDir)
  if (err) return util.cbNextTick(cb, err)

  err = util.ensureDirExists(config.influxdb.walDir)
  if (err) return util.cbNextTick(cb, err)

  const opts = {
    host: 'localhost',
    port: 10001
  }
  portFinder.getPorts(2, opts, (err, ports) =>
    gotPorts(err, context, config, ports, cb)
  )
}

// Capture the ports needed for InfluxDB.
function gotPorts (err, context, config, ports, cb) {
  if (err) return cb(err)

  config.influxdb.portCluster = ports[0]
  config.influxdb.portHTTP = ports[1]
  config.influxdb.url = `http://localhost:${ports[1]}`

  configureInfluxDB(context, config, cb)
}

// Configure InfluxDB.
function configureInfluxDB (context, config, cb) {
  const logFile = config.logs.influx

  let logFD
  try {
    logFD = fs.openSync(logFile, 'a+')
  } catch (err) {
    Logger.error(err, `error opening InfluxDB log file: ${logFile}`)
    return cb(err)
  }

  Logger.info(`writing InfluxDB log to ${logFile}`)

  const env = {
    INFLUXDB_HOSTNAME: 'localhost',
    INFLUXDB_REPORTING_DISABLED: 'true',
    INFLUXDB_DATA_INDEX_VERSION: config.influxdb.indexVersion || 'inmem',
    // following is apparently needed, but not used
    // See https://github.com/nodesource/nsolid-storage/issues/357
    INFLUXDB_BIND_ADDRESS: `localhost:${config.influxdb.portCluster}`,
    INFLUXDB_META_DIR: config.influxdb.metaDir,
    INFLUXDB_DATA_DIR: config.influxdb.dataDir,
    INFLUXDB_DATA_WAL_DIR: config.influxdb.walDir,
    INFLUXDB_ADMIN_ENABLED: false,
    INFLUXDB_HTTP_BIND_ADDRESS: `0.0.0.0:${config.influxdb.portHTTP}`,
    INFLUXDB_HTTP_LOG_ENABLED: false,
    INFLUXDB_DATA_MAX_VALUES_PER_TAG: config.influxdb.cardinalityLimit
  }

  let proc

  try {
    proc = launchInfluxDB(env, logFD)
  } catch (err) {
    Logger.error(err, 'error launching InfluxDB ...')
    return cb(err)
  }

  const influxDBProcess = new InfluxDBProcess(proc, config)

  // If the server is interrupted before the launch and bootstrap process is
  // completed, we'll still need a handle on the child process to clean it up.
  // Put this on context as a temporary name
  context.influxdbLaunchingProcess = influxDBProcess

  Logger.debug('waiting for InfluxDB to become ready ...')
  waitForInfluxDB(config, influxDBProcess, influxDBready)

  function influxDBready (err, version) {
    if (err) {
      Logger.error(`error waiting for InfluxDB to become ready: ${err}`)
      return cb(err)
    }

    createSchema(context, err => {
      if (err) {
        Logger.error(err, 'error creating InfluxDB schema')
        return cb(err)
      }
      cb(null, influxDBProcess)
    })
  }
}

function waitForInfluxDB (config, influxDBProcess, cb) {
  if (!influxDBProcess.isRunning()) {
    return cb(new Error('InfluxDB process terminated'))
  }

  if (config.influxdb.connectAttempt == null) {
    config.influxdb.connectAttempt = 0
  }

  config.influxdb.connectAttempt++
  if (config.influxdb.connectAttempt > 100) {
    return cb(new Error('unable to connect to InfluxDB'))
  }

  setTimeout(ping, 250)

  function ping () {
    Logger.debug('pinging InfluxDB to see if it\'s ready')
    api.ping(config.influxdb.url, onPing)
  }

  function onPing (err, version) {
    if (err) return waitForInfluxDB(config, influxDBProcess, cb)

    Logger.info(`InfluxDB version ${version} ready`)
    config.influxdb.version = version
    cb(null, version)
  }
}

// Launch the InfluxDB process.
function launchInfluxDB (env, logFD) {
  const platArch = `${process.platform}-${process.arch}`
  const influxPath = path.resolve(path.join(__dirname, '..', '..', 'tsdb', platArch))
  const influxExe = path.join(influxPath, 'influxd')

  env = lodash.defaults({}, env, process.env)

  const opts = {
    env: env,
    stdio: ['ignore', 'ignore', logFD]
  }

  return childProcess.spawn(influxExe, opts)
}

// This class manages the InfluxDB process.
class InfluxDBProcess {
  constructor (process, config) {
    this.process = process
    this.config = config
    this.port = config.influxdb.portHTTP
    this.url = `http://0.0.0.0:${this.port}`

    this.process.on('error', err => this._onError(err))
    this.process.on('exit', (code, signal) => this._onExit(code, signal))
  }

  isRunning () {
    return this.process != null
  }

  kill (cb) {
    if (!this.isRunning()) return

    api.shutDown()

    if (cb) this.process.on('exit', cb)
    this.process.kill()
  }

  _onError (err) {
    Logger.error(err)
    this.process = null
  }

  _onExit (code, signal) {
    Logger.info(`InfluxDB exit(${code}) signal: ${signal}`)
    this.process = null
  }
}
