'use strict'

var path = require('path')
var bole = require('bole')
var fs = require('fs')

exports.getLogger = getLogger
exports.setLogLevel = setLogLevel
exports.getLogLevel = getLogLevel

const isBrowser = typeof window === 'object' && typeof document === 'object'

let SavedLevel = 'info'
let IsIgnoringStreamError = false

let fileBuffer
if (process.env.NSOLID_CONSOLE_LOG_TO_FILE) {
  fileBuffer = fs.createWriteStream(`console-${process.pid}.log`)
}

// Stuff to do when the module is initially loaded.
function onModuleLoad () {
  setLogLevel(process.env.LOGLEVEL || process.env.LOG_LEVEL || 'info')
}

// ignore i/o errors when writing to a stream
function ignoreStreamErrorHandler (ioErr) {
  // if you want to see what errors are occuring, uncomment the following ...
  //
  // const msg = `error: ${err}\ncode:${err.code}\nstack:\n${err.stack}`
  // require('fs').writeFileSync(`error-${process.pid}.txt`, msg, {flag: 'a'})
}

// Handling of errors on stdout.
//
// It seems that sometimes at shutdown, if the node process is redirecting
// to a file, the pipe to that file is shutdown before node is finished,
// and then any subsequent writes to stdout/stderr will end up getting an
// error.
//
// So, at shutdown time only, should call `ignoreStreamError(true)` so that
// shutdown can continue, with the loss of logging, but at least the shutdown
// logic can proceed

// If passed with no arguments, return current setting.
// Otherwise set / unset the error handler.
function ignoreStreamError (bool) {
  if (bool === undefined) return IsIgnoringStreamError

  const setHandler = !!bool

  // if asking for the same setting already in use, just return
  if (setHandler === IsIgnoringStreamError) return IsIgnoringStreamError
  IsIgnoringStreamError = setHandler

  // set handler
  if (IsIgnoringStreamError) {
    process.stdout.on('error', ignoreStreamErrorHandler)
    process.stderr.on('error', ignoreStreamErrorHandler)
    return IsIgnoringStreamError
  }

  // remove handler
  process.stdout.removeListener('error', ignoreStreamErrorHandler)
  process.stderr.removeListener('error', ignoreStreamErrorHandler)
  return IsIgnoringStreamError
}

// Return a new logger for a module, named with basenames of names passed in
// combined as a path; eg, (__dirname, __filename) -> 'lib/util.js'.
function getLogger (names) {
  let name = Array.prototype.slice.call(arguments)
    .map(name => path.basename(name))
    .join(path.sep)
  name = `nsolid-console:${name}`

  const logger = isBrowser ? console : bole(name)
  logger.mute = mute
  logger.setLogLevel = setLogLevel
  logger.getLogLevel = getLogLevel
  logger.ignoreStreamError = ignoreStreamError

  return logger
}

// Set the logging level
function setLogLevel (level) {
  if (!level.match(/^(debug)|(info)|(warn)|(error)$/)) {
    level = 'info'
  }

  SavedLevel = level

  reset(SavedLevel)
}

function getLogLevel () {
  return SavedLevel
}

function mute (goSilent) {
  goSilent ? reset() : reset(SavedLevel)
}

function reset (level) {
  if (isBrowser) return

  bole.reset()

  if (level == null) return

  bole.output({
    level: SavedLevel,
    stream: fileBuffer || process.stdout
  })
}

// Run initial setup.
onModuleLoad()
