'use strict'

const snapdiff = require('snapdiff')

exports.addSummaryAsset = addSummaryAsset

const fs = require('fs')
const path = require('path')
const childProcess = require('child_process')

const bl = require('bl')
const uuid = require('uuid')
const async = require('async')

const Asset = require('./asset')
const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)

const AssetType = 'snapshot-summary'
const MAX_SNAPDIFF_MINUTES = 5
const MAX_SNAPDIFF_WORKERS = 1

const SnapDiffQueue = async.queue(snapDiffQueueProcessor, MAX_SNAPDIFF_WORKERS)

function addSummaryAsset (context, asset, cb) {
  const assetsDir = path.join(context.config.data.dir, 'assets')
  Logger.debug('adding a snapshot summary for', asset)

  createSummaryFile(asset, assetsDir, (err, summaryContent, heapTotal) => {
    if (err) {
      Logger.error(err, 'error creating snapshot summary')
      return cb(err)
    }

    // This data structure needs the same shape as the JSON header that comes
    // in over zmq, because it's passed to the same interface.
    const params = {
      bulkId: uuid.v4(),
      agentId: asset.agentId,
      type: AssetType,
      metadata: {
        viewId: asset.viewId,
        viewName: asset.viewName,
        reason: asset.reason
      }
    }

    const zmqAsset = Asset.create(context, params, (err) => {
      if (err) {
        Logger.error(err, 'error creating asset for snapshot summary')
        return cb(err)
      }

      zmqAsset.snapshotAsset = asset.bulkId
      zmqAsset.heapTotal = heapTotal

      context.diskAssetManager.add(zmqAsset)
      cb(null, zmqAsset)
    })

    zmqAsset.addData({complete: true}, summaryContent)
  })
}

function createSummaryFile (asset, assetsDir, cb) {
  const absFileName = path.join(assetsDir, `${asset.assetId}.gz`)

  createAggregate(absFileName, (err, jsonString) => {
    if (err) {
      Logger.warn(err, `error aggregating snapshot ${asset.bulkId}, file ${absFileName}`)
      return cb(err)
    }

    let jsonObject
    try {
      jsonObject = JSON.parse(jsonString)
    } catch (err) {
      Logger.warn(err, 'error parsing snapshot aggregate')
      return cb(new Error(`error parsing snapshot aggregate: ${err}`))
    }

    const fileContent = JSON.stringify({
      snapshotAsset: asset.bulkId,
      summary: jsonObject
    })

    let heapTotal = 0
    jsonObject.aggregate.forEach(item => { heapTotal += item.shallowSize })

    cb(null, fileContent, heapTotal)
  })
}

// return the result of spawning snapdiff as an errback (err, jsonString)
function createAggregate (fileName, cb) {
  SnapDiffQueue.push({fileName: fileName}, cb)
}

// ends up being called by createAggregate(), via the async queue
function snapDiffQueueProcessor (task, cb) {
  spawnSnapDiff(task.fileName, (err, result) => {
    if (err) return cb(err)
    if (result == null) return cb(new Error('no result in createAggregate() callback'))

    // return successfully
    if (result.stdout != null) return cb(null, result.stdout)

    // return with an error
    const stderr = result.stderr
    if (stderr == null) return cb(new Error('expecting stderr content but got none'))

    if (stderr.error && stderr.error.message) {
      err = new Error(stderr.error.message)
      err.data = stderr
    } else {
      err = new Error(`unknown error processing aggregate: ${JSON.stringify(stderr)}`)
    }
    return cb(err)
  })
}

// successful result from errback is {statusCode, stdout, stderr}
function spawnSnapDiff (fileName, cb) {
  const args = [
    __filename,
    fileName
  ]

  const env = Object.assign({}, process.env)
  env.NSOLID_APP = 'nsolid-snapdiff'

  const opts = {
    env: env,
    stdio: ['ignore', 'pipe', 'pipe']
  }

  Logger.debug(`spawning snapdiff with ${process.execPath} ${args.join(' ')}`)
  const proc = childProcess.spawn(process.execPath, args, opts)
  const timeout = setTimeout(timedOut, MAX_SNAPDIFF_MINUTES * 60 * 1000)
  timeout.unref()

  proc.on('close', collectResults)
  proc.on('error', finish)

  let stdoutText = null
  proc.stdout.pipe(bl((err, data) => {
    if (err) return finish(err)
    stdoutText = data.toString()
  }))

  let stderrText = null
  proc.stderr.pipe(bl((err, data) => {
    if (err) return finish(err)
    stderrText = data.toString()
  }))

  function timedOut () {
    try { proc.kill() } catch (err) {}
    finish(new Error(`snapdiff did not complete in ${MAX_SNAPDIFF_MINUTES} minutes`))
  }

  function collectResults (statusCode) {
    finish(null, statusCode, stdoutText, stderrText)
  }

  function finish (err, statusCode, stdout, stderr) {
    // clear the timeout
    clearTimeout(timeout)

    // only call the cb once!
    if (cb == null) return
    const theCB = cb
    cb = null

    if (stdout === '') stdout = null
    if (stderr === '') stderr = null

    // call the cb appropriately
    if (err != null) return theCB(err)
    theCB(null, {
      statusCode: statusCode,
      stdout: stdout,
      stderr: stderr
    })
  }
}

// we'll spawn a node process with this module to do the snapdiff work
if (require.main === module) runMain()

// entrypoint for spawned process:
//    args:   [snapshot file]
//    stdin:  not read
//    stdout: JSON summary data
//    stderr: JSON error { error: { message: 'short message'}}
function runMain () {
  process.title = 'nsolid-snapdiff'

  let fileName = process.argv[2]
  if (fileName == null) return mainExitError('no snapshot file specified')

  // ensure file name is absolute
  fileName = path.resolve(fileName)

  // check to see if file name exists
  let exists
  try {
    exists = fs.existsSync(fileName)
  } catch (err) {
    return mainExitError(`error checking snapshot file "${fileName}"`, err)
  }

  if (!exists) return mainExitError(`snapshot file does not exist: "${fileName}"`)

  // build snapdiff parms
  const snapDiffParms = {
    fileType: snapdiff.FileType.GZIP,
    fullPath: fileName
  }

  // expecting a .gz file for console server, otherwise use AUTO
  if (!fileName.toLowerCase().endsWith('.gz')) {
    snapDiffParms.fileType = snapdiff.FileType.AUTO
  }

  // aggregate!
  snapdiff.aggregate(snapDiffParms, (err, jsonString) => {
    if (err) {
      return mainExitError(`error aggregating snapshot file "${fileName}"`, err)
    }

    // try parsing it, just to make sure it's valid, since we may have issues
    // getting all the text back from the child process
    try {
      JSON.parse(jsonString)
    } catch (err) {
      return mainExitError(`error parsing snapshot aggregate for "${fileName}"`, err)
    }

    // write the output to stdout
    console.log(jsonString)
  })
}

// something went wrong, write a JSON error object to stderr
function mainExitError (message, err) {
  if (err != null) {
    message = `${message}: ${err}`
  }

  const result = {
    error: { message: message }
  }

  if (err != null) result.stack = err.stack

  console.error(JSON.stringify(result, null, 4))
  process.exit(1)
}
