'use strict'
/*
 *
 *  Websocket interface for relay
 *
 */

const RevocableEventEmitter = require('../RevocableEventEmitter')
const Engine = require('engine.io')
const EngineClient = require('engine.io-client')
const multibuffer = require('multibuffer')
const https = require('https')
const http = require('http')
const fs = require('fs')

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)

const TYPE_AUTH = 1
const TYPE_COMMAND = 2
const TYPE_DATA = 3
const TYPE_BULK = 4

class Forward extends RevocableEventEmitter {
  constructor (relay, url) {
    super()

    this.relay = relay
    this.url = url || this.relay.config.sockets.wsRemoteUrl
    this.sock = null
    this.connect()
  }

  connect () {
    this.close()

    this.sock = EngineClient(this.url, {
      transports: ['websocket'],
      rejectUnauthorized: false
    })
    this.sock.binaryType = 'blob'
    this.sock.on('open', () => {
      if (this.token) {
        this.sendAuth(this.token)
        this.sock.on('authenticated', () => {
          // unpause
          this.emit('open')
        })
      } else {
        this.emit('open')
      }
    })
    this.sock.on('message', data => this.handleMsg(data))
    this.sock.on('error', err => {
      // Don't reconnect here because close will be called too
      var msg = (err.description && err.description.message) || err.message
      Logger.error(err, `Websocket error: ${msg}`)
    })
    this.sock.on('close', () => {
      Logger.info('Reconnecting websocket forwarder')
      this.emit('close')
      setTimeout(() => this.connect(), 500)
    })
  }

  handleMsg (data) {
    var unpacked = multibuffer.unpack(data)
    switch (unpacked[0][0]) {
      case TYPE_AUTH:
        var authenticated = unpacked[1][0]
        if (authenticated) {
          this.sock.emit('authenticated')
        } else {
          Logger.error('Authentication rejected')
          this.sock.close()
        }
        break
      case TYPE_COMMAND:
        this.relay.handleCommand(unpacked[1], unpacked[2])
        break
      default:
        Logger.error(`Unhandled websocket data type ${unpacked[0][0]}`)
        break
    }
  }

  send (packed) {
    this.sock.send(packed)
  }

  sendAuth (token) {
    var tokenBuffer = Buffer.from(token)
    var bufAry = [
      Buffer([TYPE_AUTH]),
      tokenBuffer
    ]
    var packed = multibuffer.pack(bufAry)
    this.send(packed)
  }

  sendData (data) {
    var bufAry = [
      Buffer([TYPE_DATA]),
      data
    ]
    var packed = multibuffer.pack(bufAry)
    this.send(packed)
  }

  sendBulk (message, data) {
    var bufAry = [
      Buffer([TYPE_BULK]),
      message,
      data
    ]
    var packed = multibuffer.pack(bufAry)
    this.send(packed)
  }

  close () {
    this.emit('close')
    if (this.sock) {
      // Remove 'close' handler to avoid reconnecting this socket
      this.sock.removeAllListeners('close')
      this.sock.close()
      this.sock = null
    }
  }
}
module.exports.Forward = Forward

class Listener extends RevocableEventEmitter {
  constructor (relay) {
    super()

    this.relay = relay
    this.sock = null
    this.conns = new Set()
    this.listen()
  }

  listen () {
    // Discard any existing connections and shut down the
    // server if it already exists
    this.close()

    const proto = this.relay.config.sockets.wsBindProto || 'https'
    switch (proto) {
      case 'http':
        this.server = http.createServer()
        break
      case 'https':
        const options = {
          key: fs.readFileSync(this.relay.config.web.https.key),
          cert: fs.readFileSync(this.relay.config.web.https.cert)
        }
        if (this.relay.config.web.https.secureOptionsMask) {
          options.secureOptions = this.relay.config.web.https.secureOptionsMask
        }
        if (this.relay.config.web.https.ciphers) {
          options.ciphers = this.relay.config.web.https.ciphers
        }
        this.server = https.createServer(options)
        break
      default:
        throw new Error(`Unknown wsBindProto ${proto}`)
    }

    // wsBindAddr should look like "host:port" or "port".  If the user formats
    // it as "tcp://host:port" following the ZMQ syntax, discard that portion
    var wsHost
    if (this.relay.config.sockets.wsBindAddr.indexOf(':') > -1) {
      wsHost = this.relay.config.sockets.wsBindAddr
        .replace(/^tcp:\/\//, '')
        .replace(/:\d+$/, '')
    }
    var wsPort = this.relay.config.sockets.wsBindAddr.replace(/.*:/, '')
    this.server.listen(wsPort, wsHost)

    this.sock = Engine.attach(this.server)
    this.sock.binaryType = 'blob'
    this.sock.on('connection', conn => this.newConnection(conn))
  }

  newConnection (conn) {
    this.conns.add(conn)
    conn.on('message', data => this.handleMsg(conn, data))
    conn.on('close', () => {
      this.conns.delete(conn)
    })
    if (this.auth_secret) {
      conn.authenticated = false
      conn.auth_timer = setTimeout(() => {
        Logger.error('Websocket client connection auth timeout')
        conn.close()
      }, this.auth_timeout || 5000)
    } else {
      conn.authenticated = true
    }
  }

  handleMsg (conn, data) {
    var unpacked = multibuffer.unpack(data)
    switch (unpacked[0][0]) {
      case TYPE_AUTH:
        var json = unpacked[1].toString('utf8')
        this.authorize(conn, json)
        break
      case TYPE_DATA:
        if (!conn.authenticated) {
          Logger.error('Rejecting data packet because connection is unauthenticated')
          break
        }
        this.relay.handleData(unpacked[1])
        break
      case TYPE_BULK:
        if (!conn.authenticated) {
          Logger.error('Rejecting bulk packet because connection is unauthenticated')
          break
        }
        this.relay.handleBulk(unpacked[1], unpacked[2])
        break
      default:
        Logger.error(`Unhandled websocket data type ${unpacked[0][0]}`)
        break
    }
  }

  authorize (conn, token) {
    clearTimeout(conn.auth_timer)

    // TODO: verify token
    var valid = true
    if (valid) {
      conn.authenticated = true
      conn.emit('authenticated')
      this.sendAuthenticated(conn, true)
    }
  }

  send (packed) {
    for (var sock of this.conns) {
      sock.send(packed)
    }
  }

  sendAuthenticated (conn, authed) {
    var bufAry = [
      Buffer([TYPE_AUTH]),
      Buffer([authed ? 1 : 0])
    ]
    var packed = multibuffer.pack(bufAry)
    conn.send(packed)
  }

  sendCommand (topic, message) {
    var bufAry = [
      Buffer([TYPE_COMMAND]),
      topic,
      message
    ]
    var packed = multibuffer.pack(bufAry)
    this.send(packed)
  }

  close () {
    if (this.sock) {
      this.sock.close()
      this.sock = null
    }
    if (this.conns) {
      for (var sock of this.conns) {
        sock.close()
        this.conns.delete(sock)
      }
    }
  }
}
module.exports.Listener = Listener
