# Heap Snapshot Structure

## Heap Snapshot High Level Structur

```
+-- snapshot
|     |
|     +-- meta
|     |     |
|     |     +-- node_fields                   Array.<String>
|     |     |
|     |     +-- node_types                    Array.<Array.<String>,String>
|     |     |
|     |     +-- edge_fields                   Array.<String>
|     |     |
|     |     +-- edge_types                    Array.<Array.<String>,String>
|     |     |
|     |     +-- trace_function_info_fields    Array.<String>
|     |     |
|     |     +-- trace_node_fields             Array.<String>
|     |     |
|     |     +-- sample_fields                 Array.<String>
|     |
|     +-- node_count                          Number
|     |
|     +-- edge_count                          Number
|     |
|     +-- trace_function_count                Number      // most times it's 0
|
+-- nodes                                     Array.<Number>
|
+-- edges                                     Array.<Number>
|
+-- trace_function_infos                      Array.<?>       // only seen []
|
+-- trace_tree                                Array.<?>       // only seen []
|
+-- samples                                   Array.<Number>  // only seen []
|
+-- strings                                   Array.<String>
```

## Resolving Structure

### Meta Explained

- `node_types` names node types which are encoded via their index
  - first element is special as it contains array of names of node types used in `nodes` field
- `edge_types` names edge types which are encoded via their index
  - first element is special in a similar fashion as above for `node_types`

### Fields Explained

#### Nodes

`nodes` contain 5 element groups of node entries each element encoding information about one property of the node.

- the 5 elements map to node properties as follows
  - `type | name | id | self size | edge count`
- first two encode properties as an index into meta as follows
  - **type**: name can be resolved via `meta.node_types[0][type]`
  - **name**: name can be resolved from the `strings` array via `strings[name]`
- the last three need to be interpreted as literal numbers
  - **id**:  Object ID
  - **self size**: memory taken up by the node (exluding child nodes)
  - **edge count**: count of edges pointing away from the node
- see _Resolving Nodes_ too see a `nodes` code sample

#### Edges

`edges` contain 3 element groups

- the 3 elements map to edge properties as follows
  - `type | name | node entry`
- first two encode properties as an index into meta as follows
  - **type**: name can be resolved via `meta.edge_types[0][type]`
  - **name**: name can be resolved from the `strings` array via `strings[name]`
- **node entry**: the beginning of the group entry representing the node that the edge points to

#### Nodes + Edges Relationship

- the relationship of nodes and edges is encoded via their ordering
- typically the first 4 nodes are root nodes
- the first edge will point to the first node that is not root node
- that node's edge count then tells us how many edgeds are pointing away from it
- thus the edges show us the parent/child relationship of nodes in a depth first fashion

## Code Samples

### Resolving Nodes

```js
const nodes = []
for (let i = 0; i < s.nodes.length / 6; i++) {
  const j = 6 * i
  const o = {
    idx            : i,
    type           : s.nodes[j],
    typeName       : s.snapshot.meta.node_types[0][s.nodes[j]],
    name           : s.nodes[j + 1],
    nameStr        : s.strings[s.nodes[j + 1]],
    id             : s.nodes[j + 2],
    self_size      : s.nodes[j + 3],
    edge_count     : s.nodes[j + 4],
    incoming_nodes : new Set(),
    tnc            : 0
  }
  nodes.push(o)
}
```

### Resolving Edges


```js
const edges = []
for (let i = 0; i < s.edges.length / 3; i++) {
  const j = 3 * i
  const o = {
    type     : s.edges[j],
    typeName : s.snapshot.meta.edge_types[0][s.edges[j]],
    name     : s.edges[j + 1],
    nameStr  : s.strings[s.edges[j + 1]],
    to_node  : s.edges[j + 2] / 6
  }
  edges.push(o)
}
```
