#!/usr/bin/env node
'use strict'

process.title = 'nsolid-cli'
const rc = require('rc')
const minimist = require('minimist')
const Client = require('./lib/client')
const formats = require('./cli_formats')
const packageJSON = require('./package.json')

var argv = minimist(process.argv.slice(2), { 'boolean': ['delete'] })
var conf = rc('nsolid-cli', {
  remote: 'http://localhost:6753/api/'
}, argv)

if (!/^http/.test(conf.remote)) {
  conf.remote = 'http://' + conf.remote
}

if (!/\/api\/$/.test(conf.remote)) {
  conf.remote = conf.remote + '/api/'
}

if (conf.h || conf.help) {
  const helpFile = require('path').join(__dirname, 'help.txt')
  const helpText = require('fs').readFileSync(helpFile)
  console.error(String(helpText))
  process.exit(0)
}

if (conf.s || conf.stdin) {
  // passing -s or --stdin expects to read from stdin
  conf.body = process.stdin
}

if (conf.v || conf.version) {
  console.log(packageJSON.version)
  process.exit(0)
}

if (conf._.length === 0) {
  usageError('Expecting a command: snapshot, etc. Use -h for list of commands.')
}

if (conf._.length > 1) {
  usageError('Expecting only a single command')
}

var command = conf._[0]
// 'ls' is an alias for 'list'
if (command === 'ls') {
  command = 'list'
}

const client = new Client(conf.remote, function onReady (err) {
  if (err) {
    // Error setting up client, bad server?
    console.error('Failed to get schema from API server:')
    return usageError(err.message)
  }
})

if (command === 'snapshot' || command === 'profile') {
  // trigger
  client.run(command, conf, (err, res, reply) => {
    if (err) {
      return usageError(err.message)
    }
    try {
      const payload = JSON.parse(reply)
      if (payload && payload.asset) {
        // fetch result
        const assetConf = {
          id: payload.asset,
          auth: conf.auth
        }
        client.run('asset', assetConf, (err, res, reply) => {
          if (err) {
            return usageError(err.message)
          }
          process.stdout.write(reply)
        })
      } else {
        console.error('Unable to fetch created asset!')
        console.error(payload)
      }
    } catch (e) {
      console.error(e)
    }
  })
} else {
  client.run(command, conf, (err, res, reply) => {
    if (err) {
      return usageError(err.message)
    }
    const printTo = /^2/.test(res.statusCode) ? process.stdout : process.stderr
    // reply can be a reply OR a stream
    if (reply.pipe != null && (typeof reply.pipe === 'function')) {
      if (conf.fields != null && command !== 'asset') {
        return reply.pipe(formats.filterFields(conf.fields)).pipe(formats.Consolize(printTo)).pipe(printTo)
      }
      return reply.pipe(formats.Consolize(printTo)).pipe(printTo)
    }
    printTo.write(reply)
    if (printTo.isTTY) {
      // append a newline if it didn't end with one
      if (reply[reply.length - 1] !== 10) {
        printTo.write('\n')
      }
    }
  })
}

function usageError (message) {
  console.error(message)
  process.exit(1)
}
